package com.vertigrated.oss.jackson.extjs.treestore;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.google.common.collect.ImmutableSet;
import com.vertigrated.oss.extjs.treestore.ContainerNode;
import com.vertigrated.oss.extjs.treestore.ObjectNode;
import com.vertigrated.oss.extjs.treestore.ValueNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.*;

public class ObjectNodeSerializer extends JsonSerializer<ObjectNode>
{
    private static final Logger L = LoggerFactory.getLogger(ObjectNodeSerializer.class);

    private static final Set<Class<?>> VALUE_TYPES;
    private static final Set<Class<?>> CONTAINER_TYPES;

    static
    {
        VALUE_TYPES = ImmutableSet.<Class<?>>of(String.class, Integer.class, Double.class, Float.class, int.class, double.class, float.class,
                                                Boolean.class, boolean.class, Date.class);
        CONTAINER_TYPES = ImmutableSet.<Class<?>>of(List.class, Set.class);
    }

    @Override
    public void serialize(ObjectNode value, JsonGenerator jgen, SerializerProvider serializers) throws IOException
    {
        final Object object = value.object;
        final Field[] declaredFields = object.getClass().getDeclaredFields();
        final List<Field> fields = Arrays.asList(declaredFields);
        final Set<Field> values = this.getFields(fields);
        if (values.isEmpty())
        {
            jgen.writeObject(new ValueNode(value.id, value.name, object.getClass(), value.cls, object));
        }
        else
        {
            jgen.writeStartObject();
            final String parent = value.id + "/" + object.getClass().getSimpleName();
            jgen.writeStringField("id", parent);
            jgen.writeStringField("name", object.getClass().getSimpleName());
            jgen.writeBooleanField("leaf", false);
            jgen.writeArrayFieldStart("children");
            for (final Field f : values)
            {
                final Object fieldValue = this.getFieldValue(f, object);
                if (VALUE_TYPES.contains(f.getType()))
                {
                    final ValueNode ln = new ValueNode(parent + "/" + f.getName(), f.getName(), f.getType(), "value", fieldValue);
                    jgen.writeObject(ln);
                }
                else if (CONTAINER_TYPES.contains(f.getType()))
                {
                    final ParameterizedType pt = ParameterizedType.class.cast(f.getGenericType());
                    final Class<?> type = (Class<?>) pt.getActualTypeArguments()[0];
                    final ContainerNode nln = new ContainerNode(parent + "/" + f.getName(), f.getName(), f.getType(), type, "list", (Collection) fieldValue);
                    jgen.writeObject(nln);
                }
                else
                {
                    final Object fv = this.getFieldValue(f, object);
                    if (fv != null)
                    {
                        final ObjectNode on = new ObjectNode(parent + "/" + f.getName(), f.getName(), f.getType(), "object", fv);
                        jgen.writeObject(on);
                    }
                }
            }
            jgen.writeEndArray();
            jgen.writeEndObject();
        }
    }

    private Set<Field> getFields(List<Field> fields)
    {
        final ImmutableSet.Builder isb = ImmutableSet.<Field>builder();
        for (final Field field : fields)
        {
            field.setAccessible(true);
            if (isJsonProperty(field))
            {
                isb.add(field);
            }
        }
        return isb.build();
    }

    private boolean isJsonProperty(@Nonnull final Field f) { return f.getAnnotation(JsonProperty.class) != null; }

    private Object getFieldValue(@Nonnull final Field f, @Nonnull Object value)
    {
        try
        {
            f.setAccessible(true);
            return f.get(value);
        }
        catch (IllegalAccessException e)
        {
            throw new RuntimeException(e);
        }
    }
}
