package com.vertigrated.oss.jackson.jdbc;

import static java.sql.ResultSetMetaData.columnNullable;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import javax.annotation.Nonnull;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;

public class ResultSetSerializer extends JsonSerializer<ResultSet>
{
    public enum ColumnFeatures
    {
        ALL,
        INDEX_ONLY,
        NAME_ONLY,
        INDEX_NAME_ONLY
    }

    public enum RowFeatures
    {
        NO_INDEX_OR_NAME,
        NAME_ONLY,
        INDEX_ONLY,
        INDEX_AND_NAME
    }

    private final ColumnFeatures columnFeatures;
    private final RowFeatures rowFeatures;

    public ResultSetSerializer()
    {
        this(ColumnFeatures.ALL, RowFeatures.NO_INDEX_OR_NAME);
    }

    public ResultSetSerializer(@Nonnull final ColumnFeatures columnFeatures)
    {
        this(columnFeatures, RowFeatures.NO_INDEX_OR_NAME);
    }

    public ResultSetSerializer(@Nonnull final RowFeatures rowFeatures)
    {
        this(ColumnFeatures.ALL, rowFeatures);
    }

    public ResultSetSerializer(@Nonnull final ColumnFeatures columnFeatures, final RowFeatures rowFeatures)
    {
        this.columnFeatures = columnFeatures;
        this.rowFeatures = rowFeatures;
    }

    @Override
    public void serialize(@Nonnull final ResultSet rs, @Nonnull final JsonGenerator jgen, @Nonnull final SerializerProvider provider) throws IOException, JsonProcessingException
    {
        try
        {
            final ResultSetMetaData rsm = rs.getMetaData();
            final int columnCount = rsm.getColumnCount();
            jgen.writeStartObject();

            jgen.writeArrayFieldStart("columns");
            for (int i = 1; i <= columnCount; i++)
            {
                jgen.writeStartObject();
                if (this.columnFeatures.equals(ColumnFeatures.ALL) ||
                        this.columnFeatures.equals(ColumnFeatures.INDEX_NAME_ONLY) ||
                        this.columnFeatures.equals(ColumnFeatures.INDEX_ONLY))
                {
                    jgen.writeNumberField("index", i);
                }
                if (this.columnFeatures.equals(ColumnFeatures.ALL) ||
                        this.columnFeatures.equals(ColumnFeatures.NAME_ONLY) ||
                        this.columnFeatures.equals(ColumnFeatures.INDEX_NAME_ONLY))
                {
                    jgen.writeStringField("name", rsm.getColumnName(i));
                }
                if (this.columnFeatures.equals(ColumnFeatures.ALL))
                {
                    jgen.writeStringField("type", rsm.getColumnTypeName(i));
                    jgen.writeNumberField("width", rsm.getColumnDisplaySize(i));
                    jgen.writeBooleanField("nullable", rsm.isNullable(i) == columnNullable);
                }
                jgen.writeEndObject();
            }
            jgen.writeEndArray();

            jgen.writeArrayFieldStart("rows");
            while (rs.next())
            {
                for (int i = 1; i <= columnCount; i++)
                {
                    if (this.rowFeatures.equals(RowFeatures.NO_INDEX_OR_NAME))
                    {
                        jgen.writeString(rs.getString(i));
                    }
                    else
                    {
                        jgen.writeStartObject();
                        if (this.rowFeatures.equals(RowFeatures.INDEX_ONLY) ||
                                this.rowFeatures.equals(RowFeatures.INDEX_AND_NAME))
                        {
                            jgen.writeNumberField("index", i);
                        }
                        if (this.rowFeatures.equals(RowFeatures.NAME_ONLY) ||
                                this.rowFeatures.equals(RowFeatures.INDEX_AND_NAME))
                        {
                            jgen.writeStringField("name", rsm.getColumnName(i));
                        }
                        jgen.writeStringField("value", rs.getString(i));
                        jgen.writeEndObject();
                    }
                }
            }
            jgen.writeEndArray();

            jgen.writeEndObject();
        }
        catch (final SQLException e)
        {
            throw new RuntimeException(e);
        }
    }

    @Nonnull
    @Override
    public Class<ResultSet> handledType()
    {
        return ResultSet.class;
    }
}
